/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	cubetest.cpp

	This small sample application shows how to read a block from
	an extent file. The resulting sequence of planes is stored to 
	a bmp image.
*/

#include <windows.h>
#include "../ExtentFile/ExtentFile.h"

using namespace extfile;

//! Write a 32bpp Windows .bmp file
/*!
\param fname File name
\param data Bitmap data
\param w Bitmap width
\param h Bitmap height
*/
void writeBitmap(const char *fname, void *data, UInt32 w, UInt32 h)
{
	// Open the file
	FILE *of=fopen(fname,"wb");
	if(of==NULL)
		return;

	// Fill out bitmap header structures
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	bfh.bfType='B'+('M'<<8);
	bfh.bfOffBits=sizeof(bfh)+sizeof(bih);
	bfh.bfSize=sizeof(bfh)+sizeof(bih)+w*h*4;
	bfh.bfReserved1=0;
	bfh.bfReserved2=0;
	bih.biBitCount=32;
	bih.biClrImportant=0;
	bih.biClrUsed=0;
	bih.biCompression=0;
	bih.biHeight=h;
	bih.biPlanes=1;
	bih.biSize=sizeof(bih);
	bih.biSizeImage=w*h*4;
	bih.biWidth=w;
	bih.biXPelsPerMeter=0;
	bih.biYPelsPerMeter=0;

	// Write it all to the file
	fwrite(&bfh,1,sizeof(bfh),of);
	fwrite(&bih,1,sizeof(bih),of);
	// The bitmap must be flipped vertically
	for(UInt32 i=0;i<h;i++)
		fwrite(((UInt32*)data)+(h-1-i)*w,w,4,of);

	// Close the file
	fclose(of);
}

//! Main function
void main(int argc,	char *argv[])
{
	if(argc==1)
	{
		printf("Usage: convert <xmlfile1> ...\n");
		return;
	}

	//////////////////////////////////////////////////////////////
	//
	// Sample 1: Reading cubes from an extent file
	//
	//////////////////////////////////////////////////////////////

	ExtentFile ef;

	// Open the extent file
	if(!ef.open(argv[1]))
	{
		printf("Could not open %s\n",argv[1]);
		return;
	}

	// Allocate some memory for output
	UInt8 *cube=(UInt8*)malloc(256*256*4*4);

	// Get a cube (256x256x4 voxels) from the dataset 
	// Resolution: 0 (maximum)
	// Start coordinates: 128,128,100
	// End coordinates: 384,384,104
	// Channel: 0
	// Output buffer: cube
	ef.getCube(0,128,128,100,384,384,104,0,cube);

	// Now save the data to a bitmap
	writeBitmap("dummy.bmp",cube,256,1024);

	// Close the extent file
	ef.close();

	//////////////////////////////////////////////////////////////
	//
	// Sample 2: Reading individual voxels from a cached extent file
	//
	//////////////////////////////////////////////////////////////

	CachedExtentFile ef2;

	// Open the extent file, allocate 32MB cache
	if(!ef2.open(argv[1],32*1048576))
	{
		printf("Could not open %s\n",argv[1]);
		return;
	}

	UInt32 *slice=(UInt32*)malloc(1024*1024*4);

	// Get a slice from the dataset
	for(int j=0;j<256;j++)
	{
		for(int i=0;i<256;i++)
		{
			// Get a voxel from the dataset
			// Resolution: 0 (maximum)
			// Voxel coordinates: 128+i,128+j,100
			// Channel: 0
			slice[i+j*256]=ef2.getVoxel(0,128+i,128+j,100,0);
		}
	}

	// Now save the data to a bitmap
	writeBitmap("dummy2.bmp",slice,256,256);

	//////////////////////////////////////////////////////////////
	//
	// Sample 3: Reading a slice from a cached extent file
	//
	//////////////////////////////////////////////////////////////

	// Extract a slice from the dataset
	// Slice origin: 80,80,100
	// Slice right vector: 1,0,0
	// Slice up vector: 0,0,1
	// Slice zoom factor: 3
	// Slice output bitmap size: 512 512
	// Slice output buffer: slice
	// Slice channel: 0
	ef2.getSlice(80,80,100,1,0,0,0,0,1,3,512,512,slice,0);

	// Now save the data to a bitmap
	writeBitmap("dummy3.bmp",slice,1024,1024);

	// Close the extent file
	ef2.close();

}