/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __EXTENTFILE_H
#define __EXTENTFILE_H

#include "ExtentFileDescriptor.h"
#include "File.h"

//! Namespace for extent file
namespace extfile
{
	UInt32 defaultStripe(UInt32 in, UInt32 sc);


	//! A dataset label
	/*!
	Labels are used for segmenting a dataset into individual structures. The
	label structure here is specific to the Visible Human dataset used on the
	EPFL web server (http://visiblehuman.epfl.ch).
	*/
	struct LabelDescription
	{
		//! Name of the structure
		char name[64];
		//! Side of the body (L or R)
		char side[2];
		//! System to which the structure belongs (Nervous, Muscular, Skeletal, etc.)
		char system[32];
		//! Location of the structure (Head, Torso, etc.)
		char location[32];
		//! Indicates whether this label is used
		bool use;
		//! Should this label be displayed
		bool display;
	};

	//! Class for reading extent files
	/*!
	An extent file is a representation of a volume dataset on disk where the data is
	segmented into small cubes. This class is intended for applications that require
	low level access to extent files, since it exposes only two read methods: one to
	read a given subvolume of an extent file, and another to read a single extent.
	For more functions, you can used CachedExtentFile.
	*/
	class ExtentFile
	{
	public:
		//! Descriptor of the extent file
		ExtentFileDescriptor efd;

	public:
		//! Number of stripes
		Int32 diskCount;
		//! File handles for the stripes
		FileWrapper data[MAXSTRIPES];

		//! Address table
		UInt32 *address;

		//! Temp buffer for decoding extents
		UInt8 *tempBuffer;


	public:
		//! Number of resolutions
		Int32 zoomLevels;
		//! Offset of the first extent for each resolution
		UInt32 extOff[MAXZOOM];
		Int32 extX[MAXZOOM], extY[MAXZOOM], extZ[MAXZOOM];
		Int32 sizeX[MAXZOOM], sizeY[MAXZOOM], sizeZ[MAXZOOM];
		//! Size of the extents (including fat)
		Int32 extSizeX, extSizeY, extSizeZ;
		//! Useful size of extents (excluding fat)
		Int32 realExtSizeX, realExtSizeY, realExtSizeZ;
		//! Complete size of dataset at highest resolution
		Int32 imgSizeX, imgSizeY, imgSizeZ;
		//! Total number of extents
		Int32 extCount;
		//! Size of an extent in bytes
		Int32 extSize;

		//! Size of voxels
		double voxelSizeX, voxelSizeY, voxelSizeZ;

		//! Labels
		LabelDescription *labels;
		//! Number of labels
		Int32 labelCount;

		//! Is this file open?
		bool active;

		//! Striping function
		StripeFunc stripefunc;

	private:
		//! Load labels from a label file
		bool loadLabels(const char *fname);

	public:
		//! Constructor
		ExtentFile() { active = false; stripefunc=defaultStripe; labels=NULL; labelCount=0; }
		//! Destructor
		~ExtentFile() { if(active) close(); }

		//! Open the extent file
		bool open(const char *configFileName);
		//! Close the extent file
		bool close();

		//! Load an extent
		/*!
		\param en index of the extent to load
		\param target target buffer for extent
		The target buffer must be at least the size of the extent (complete size of extent
		taking into account all fat and all channels, rounded up to the next extent file
		block size)
		*/
		void loadExtent(int en, void *target);

		//! Load a subset of the volume
		/*!
		\param r resolution
		\param x0 X coordinate of upper left corner
		\param y0 Y coordinate of upper left corner
		\param z0 Z coordinate of upper left corner
		\param x1 X coordinate of lower right corner
		\param y1 Y coordinate of lower right corner
		\param z1 Z coordinate of lower right corner
		\param c channel to read
		\param target target buffer for extent
		*/
		bool getCube(int r, int x0, int y0, int z0, int x1, int y1, int z1, int c, void *target);
	};


	//! Linked list for LRU extent cache
	class LoadedExtent
	{
	public:
		//! Currently occupied cache slot
		Int32 slot;
		//! Number of extent
		Int32 extent;
		//! Address of extent in extent cache
		UInt32 *address;
		//! Real address of extent (might not be in cache), is set when extent is loaded
		UInt32 *finalAddress;
		//! Previous pointer for linked list 
		LoadedExtent *prev;
		//! Next pointer for linked list 
		LoadedExtent *next;
	};

	//! Extent file with extent cache
	/*!
	This class provides read functionality from extent files enabling access to
	individual voxels. To keep this somewhat efficient, the extents are stored
	in a cache, with a least recently used mechanism to dump extents when more
	space is required.
	*/
	class CachedExtentFile : public ExtentFile
	{
		//! Size of cache
		UInt32 cacheSize;
		//! List of extents
		LoadedExtent *lruListItem;
		//! Base address of cache
		UInt8 *cacheBase;
		//! Address of lowest resolution extents
		UInt8 *bottom;
		//! List of currently loaded extents
		LoadedExtent *cacheHead, *cacheTail;
		//! Number of free cache slots
		int unusedCacheSlots;

		//! Is this extent file open?
		bool active;

		//! Zero filled buffer for empty extents
		UInt32 *zeroBuffer;
		//! Offset of channels
		Int32 baseSize[MAXCHANNELS];

	private:
		//! Load a specific extent into the cache
		bool loadExtent(int en);

		//! Get a pointer to a specific channel in an extent
		/*!
		\param en Index of extent in extent file
		\param c Channel
		*/
		void *getExtent(int en, int c) 
		{ 
			if(en<0 || en>=extCount || c<0 || c>=efd.channels)
				return zeroBuffer;
			if(!lruListItem[en].finalAddress) 
				loadExtent(en); 
			return ((UInt8*)lruListItem[en].finalAddress)+baseSize[c]; 
		}

	public:
		//! Constructor
		CachedExtentFile() { active=false; }
		//! Destructor
		~CachedExtentFile() { if(active) close(); }

		//! Open an extent file
		/*!
		\param configFileName Name of XML file containing extent file description
		\param cacheSize Size of cache in bytes
		*/
		bool open(const char *configFileName, const UInt32 cacheSize);
		//! Close an extent file
		bool close();

		//! Get a pointer to a specific channel in an extent
		/*!
		\param rf Requested resolution
		\param x X position of extent in dataset
		\param y Y position of extent in dataset
		\param z Z position of extent in dataset
		\param c Channel
		*/
		void *getExtent(int rf, int x, int y, int z, int c) 
		{ 
			if(rf<0 || rf>=zoomLevels || x<0 || y<0 || z<0 || x>=extX[rf] || y>=extY[rf] || z>=extZ[rf] || c<0 || c>=efd.channels) 
				return zeroBuffer; 
			return ((UInt8*)getExtent(x+(y*extX[rf])+(z*extX[rf]*extY[rf])+extOff[rf],c)); 
		}

		//! Get a specific voxel from the dataset
		/*!
		\param rf Requested resolution
		\param x X position of voxel in dataset
		\param y Y position of voxel in dataset
		\param z Z position of voxel in dataset
		\param c Channel
		*/
		UInt32 getVoxel(int rf, int x, int y, int z, int c) 
		{ 
			if(rf<0 || rf>=zoomLevels || x<0 || y<0 || z<0 || x>=sizeX[rf] || y>=sizeY[rf] || z>=sizeZ[rf] || c<0 || c>=efd.channels) 
				return 0; 
			UInt8 *off=((UInt8*)getExtent(rf,x/realExtSizeX,y/realExtSizeY,z/realExtSizeZ,c))+((x%realExtSizeX)+((y%realExtSizeY)*realExtSizeX)+((z%realExtSizeZ)*realExtSizeX*realExtSizeY))*efd.channelSize(efd.channelFormat[c]);
			UInt32 rv=0;
			for(int i=0;i<efd.channelSize(efd.channelFormat[c]);i++)
				rv|=off[i]<<(i<<3);
			return rv;
		}

		//! Extract a slice from the dataset
		/*!
		\param xp X position of center of slice
		\param yp Y position of center of slice
		\param zp Z position of center of slice
		\param ux X component of right vector in slice
		\param uy Y component of right vector in slice
		\param uz Z component of right vector in slice
		\param vx X component of up vector in slice
		\param vy Y component of up vector in slice
		\param vz Z component of up vector in slice
		\param zoom Zoom factor for slice (in pixels per unit)
		\param w Width of output bitmap
		\param h Height of output bitmap
		\param c Channel to extract slice from
		This method extracts slices of arbitrary orientation from the dataset. It is
		provided to illustrate one method for slice extraction, and is not optimized
		for performance. The units for position and zoom are given by the aspect ratio
		specified in the XML descriptor.
		*/
		void getSlice(float xp, float yp, float zp, float ux, float uy, float uz, float vx, float vy, float vz, float zoom, int w, int h, void *data, int c);

	};

}

#endif