/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	ExtentFileDescriptor.cpp

	This is the implementation of the ExtentFileDescriptor class
*/

#include "ExtentFileDescriptor.h"

namespace extfile
{

	static Int32 channelFromString(const char *str)
	{
		// Convert a string to an equivalent identifier
		if(!strcmp(str,"R8G8B8"))
			return ExtentFileDescriptor::R8G8B8;
		else if(!strcmp(str,"X8R8G8B8"))
			return ExtentFileDescriptor::X8R8G8B8;
		else if(!strcmp(str,"B8G8R8X8"))
			return ExtentFileDescriptor::B8G8R8X8;
		else if(!strcmp(str,"R5G6B5"))
			return ExtentFileDescriptor::R5G6B5;
		else if(!strcmp(str,"L16"))
			return ExtentFileDescriptor::L16;
		else if(!strcmp(str,"L16_BIG"))
			return ExtentFileDescriptor::L16_BIG;
		else if(!strcmp(str,"G8"))
			return ExtentFileDescriptor::G8;
		else if(!strcmp(str,"G16"))
			return ExtentFileDescriptor::G16;
		return ExtentFileDescriptor::INVALID;
	}

	Int32 Vector<Int32>::convert(const char *a)
	{
		// Convert string to int
		return atoi(a);
	}

	double Vector<double>::convert(const char *a)
	{
		// Convert string to double
		return atof(a);
	}

	bool Stripe::loadFromXML(xmlDocPtr doc, xmlNodePtr cur)
	{
		// Standard XML parser stuff
		cur=cur->xmlChildrenNode;
		while(cur!=NULL) 
		{
			if ((!xmlStrcmp(cur->name, (const xmlChar *)"host")))
				host=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"path")))
				path=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"share")))
				share=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"index")))
				index=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"data")))
				data=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"segment")))
				segment=_atoi64((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			cur = cur->next;
		}
		return true;
	}

	bool FilePattern::loadFromXML(xmlDocPtr doc, xmlNodePtr cur)
	{
		// Standard XML parser stuff
		cur=cur->xmlChildrenNode;
		while(cur!=NULL) 
		{
			if ((!xmlStrcmp(cur->name, (const xmlChar *)"name")))
				name=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			if ((!xmlStrcmp(cur->name, (const xmlChar *)"imageformat")))
				format=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"width")))
				width=atoi((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"height")))
				height=atoi((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"count")))
				count=atoi((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"offset")))
				offset=atoi((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			else if ((!xmlStrcmp(cur->name, (const xmlChar *)"format")))
				channel=channelFromString((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			cur = cur->next;
		}
		return true;
	}

	bool ExtentFileDescriptor::loadFromXML(const char *filename)
	{
		// Standard XML parser stuff
		LIBXML_TEST_VERSION
			xmlKeepBlanksDefault(0);

		xmlDocPtr doc;
		xmlNodePtr cur;

		doc=xmlParseFile(filename);
		if(doc==NULL) 
			return false;

		cur=xmlDocGetRootElement(doc);
		if(cur==NULL) 
		{
			fprintf(stderr,"empty document\n");
			xmlFreeDoc(doc);
			return false;
		}
		if(xmlStrcmp(cur->name,(const xmlChar *)"dataset")) 
		{
			fprintf(stderr,"document of the wrong type, root node != dataset");
			xmlFreeDoc(doc);
			return false;
		}

		cur=cur->xmlChildrenNode;
		while(cur && xmlIsBlankNode(cur))
			cur=cur->next;

		while (cur != NULL) 
		{
			if(!xmlStrcmp(cur->name,(const xmlChar *)"name")) 
				name=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"comment")) 
				comment=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"origin")) 
				origin=(char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"resolutions")) 
				resolutions=atoi((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"size")) 
				datasetSize.loadFromXML(doc,cur);
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"aspect")) 
				voxelSize.loadFromXML(doc,cur);
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"extentsize")) 
			{
				extentSize.loadFromXML(doc,cur);
				char *flgs=(char *)xmlGetProp(cur,(const xmlChar *)"flags");
				if(flgs && strstr(flgs,"fatx"))
					flags|=FATX;
				if(flgs && strstr(flgs,"faty"))
					flags|=FATY;
				if(flgs && strstr(flgs,"fatz"))
					flags|=FATZ;
			}
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"stripes")) 
			{
				xmlNodePtr lcur=cur->xmlChildrenNode;
				while(lcur!=NULL) 
				{
					if ((!xmlStrcmp(lcur->name, (const xmlChar *)"stripe")))
					{
						stripes[stripeCount++].loadFromXML(doc,lcur);
					}
					lcur=lcur->next;
				}
			}
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"format")) 
			{
				char *flgs=(char *)xmlGetProp(cur,(const xmlChar *)"flags");
				if(flgs && strstr(flgs,"compressed"))
					flags|=COMPRESSED;
				xmlNodePtr lcur=cur->xmlChildrenNode;
				blockSize=atoi((char *)xmlGetProp(cur,(const xmlChar *)"blocksize"));

				while(lcur!=NULL) 
				{
					if ((!xmlStrcmp(lcur->name, (const xmlChar *)"channel")))
					{
						char *cattr=(char *)xmlGetProp(lcur,(const xmlChar *)"index");
						char *clabel=(char *)xmlGetProp(lcur,(const xmlChar *)"labels");
						int ndx=atoi(cattr);
						if(ndx+1>channels)
							channels=ndx+1;

						xmlNodePtr lcur2=lcur->xmlChildrenNode;
						while(lcur2!=NULL) 
						{
							if ((!xmlStrcmp(lcur2->name, (const xmlChar *)"format")))
							{
								char *ctype=(char*)xmlNodeListGetString(doc, lcur2->xmlChildrenNode, 1);
								channelFormat[ndx]=channelFromString(ctype);
							}
							if ((!xmlStrcmp(lcur2->name, (const xmlChar *)"filepattern")))
							{
								sources[ndx].loadFromXML(doc,lcur2);
							}
							lcur2=lcur2->next;
						}

						if(clabel)
							labels=clabel;
					}
					lcur=lcur->next;
				}
			}
			else if(!xmlStrcmp(cur->name,(const xmlChar *)"source")) 
			{
				xmlNodePtr lcur=cur->xmlChildrenNode;
				while(lcur!=NULL) 
				{
					if ((!xmlStrcmp(lcur->name, (const xmlChar *)"filepattern")))
					{
						char *cattr=(char *)xmlGetProp(lcur,(const xmlChar *)"index");
						int ndx=atoi(cattr);
						sources[ndx].loadFromXML(doc,lcur);
					}
					lcur=lcur->next;
				}
			}
			cur = cur->next;
		}

		// Now calculate some useful stuff

		// Number of extents in all directions
		extX[0]=(datasetSize.x-1)/extentSize.x+1;
		extY[0]=(datasetSize.y-1)/extentSize.y+1;
		extZ[0]=(datasetSize.z-1)/extentSize.z+1;
		// Size of dataset rounded up to extent size
		sizeX[0]=extentSize.x*extX[0];
		sizeY[0]=extentSize.y*extY[0];
		sizeZ[0]=extentSize.z*extZ[0];

		// Now compute the same for all resolutions
		int ec=extX[0]*extY[0]*extZ[0];
		for(int i=1;i<=resolutions;i++)
		{
			extX[i]=(extX[i-1]+1)/2;
			extY[i]=(extY[i-1]+1)/2;
			extZ[i]=(extZ[i-1]+1)/2;
			sizeX[i]=extentSize.x*extX[i];
			sizeY[i]=extentSize.y*extY[i];
			sizeZ[i]=extentSize.z*extZ[i];
			ec+=extX[i]*extY[i]*extZ[i];
			extOff[i]=extOff[i-1]+extX[i-1]*extY[i-1]*extZ[i-1];
		}
		// Total number of extents
		extOff[resolutions]=ec;
		// Calculate offsets of extents in extent file
		extOff[resolutions-1]=0;
		for(i=resolutions-1;i>0;i--)
			extOff[i-1]=extOff[i]+extX[i]*extY[i]*extZ[i];

		// Full size of extent (including fat)
		fullExtentSize.x=extentSize.x+((flags&FATX) ? 1 : 0);
		fullExtentSize.y=extentSize.y+((flags&FATY) ? 1 : 0);
		fullExtentSize.z=extentSize.z+((flags&FATZ) ? 1 : 0);

		// Total bytes per pixel
		Int32 ops=0;
		for(i=0;i<channels;i++)
			ops+=channelSize(channelFormat[i]);

		// Size of an extent (exact size in bytes)
		extSize=fullExtentSize.x*fullExtentSize.y*fullExtentSize.z*ops;

		// Clean up	everything else	before quitting.
		xmlCleanupParser();

		return true;
	}

}
