/*
	Extent File System - Volume data file storage
	Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __EXTENTFILEDESC_H
#define __EXTENTFILEDESC_H

#include <string.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

typedef unsigned char UInt8;
typedef unsigned short UInt16;
typedef unsigned int UInt32;
typedef int Int32;
typedef __int64 Int64;

namespace extfile
{
	//! Type for striping function	
	typedef UInt32 (*StripeFunc)(UInt32 in, UInt32 sc);
	
	//! Maximum number of stripes
	const Int32 MAXSTRIPES = 16;
	//! Maximum number of zoom levels
	const Int32 MAXZOOM	= 16;
	//! Maximum number of source datasets (channels)
	const Int32 MAXCHANNELS	= 16;
	
#ifndef __CAP2_BASE_H
	//! Simple string wrapper class (taken from CAP2 library and trimmed slightly)
	class StringA
	{
		char *value;
	public:
		StringA() { value=NULL; }
		StringA(const char *v) { value=NULL;	if(v) { value=new char[strlen(v)+1]; strcpy(value,v); } }
		StringA(const StringA &v) { value=NULL; if(v.value) { value=new char[strlen(v.value)+1]; strcpy(value,v.value); } }
		StringA(const StringA *v) { value=NULL; if(v && v->value) { value=new char[strlen(v->value)+1]; strcpy(value,v->value); } }
		~StringA() { if(value) delete value; }
		Int32 length() { return value ? strlen(value) : 0; }
		bool isNull() { return value==NULL; }
		
		StringA& operator=(char *val) { if(value) delete value; value=NULL; if(val) { value=new char[strlen(val)+1]; strcpy(value,val);	} return *this; }
		StringA& operator=(const StringA& val) { if(value) delete value; value=NULL; if(val.value) { value=new char[strlen(val.value)+1]; strcpy(value,val.value); } return *this; }
		StringA& operator=(const StringA* val) { if(value) delete value; value=NULL; if(val && val->value) { value=new char[strlen(val->value)+1]; strcpy(value,val->value); } return *this; }
		operator const char *() { return value ? value : "(null)"; }
	};
#endif
	
	//! 3D Vector
	template<class et> struct Vector
	{
		et x;	//!< X coordinate
		et y;	//!< Y coordinate
		et z;	//!< Z coordinate
		
		//! Constructor
		Vector() { x=y=z=0; }
		//! Conversion function for vector elements (wraps atoi, atof, etc...)
		et convert(const char *a);
		//! Load vector from XML document
		bool loadFromXML(xmlDocPtr doc, xmlNodePtr cur);
	};
	
	template<class et> bool Vector<et>::loadFromXML(xmlDocPtr doc, xmlNodePtr cur)
	{
		cur=cur->xmlChildrenNode;
		while(cur!=NULL) 
		{
			if ((!xmlStrcmp(cur->name, (const xmlChar *)"x")))
				x=convert((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			if ((!xmlStrcmp(cur->name, (const xmlChar *)"y")))
				y=convert((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			if ((!xmlStrcmp(cur->name, (const xmlChar *)"z")))
				z=convert((char*)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
			cur = cur->next;
		}
		return true;
	}
	
	//! Descriptor for a stripe of the extent file
	class Stripe
	{
	public:
		StringA host;		//!< Host on which stripe resides (currently unused)
		StringA path;		//!< Local path to stripe 
		StringA share;		//!< Share name for stripe host (currently unused)
		StringA index;		//!< Index file name
		StringA data;		//!< Data file name
		Int64 segment;		//!< Maximum file segment size (currently unused)
		
		//! Load stripe descriptor from XML
		bool loadFromXML(xmlDocPtr doc, xmlNodePtr cur);
		
		//! Constructor
		Stripe()
		{
			segment=0;
		}
	};
	
	//! Descriptor for an input file pattern
	class FilePattern
	{
	public:
		StringA name;	//!< File name
		StringA format;	//!< Image format
		Int32 channel;	//!< Channel index in resulting dataset
		Int32 width;	//!< Width of images
		Int32 height;	//!< Height of images
		Int32 offset;	//!< Offset of first file
		Int32 count;	//!< Number of files

		//! Constructor
		FilePattern()
		{
			channel=0; width=0; height=0; offset=0; count=0;
		}
		
		//! Load descriptor from XML
		bool loadFromXML(xmlDocPtr doc, xmlNodePtr cur);
	};
	
	//! Entry in address table file
	struct AddressEntry
	{
		Int32 offset;	//!< Offset in data file (in blocks)
		Int32 length;	//!< Length of data (in bytes)
	};
	
	//! Descriptor for an extent file
	class ExtentFileDescriptor
	{
	public:
		enum { 
			FATX		= 0x00000001,	//!< Fat extents in X
			FATY		= 0x00000002,	//!< Fat extents in Y
			FATZ		= 0x00000004,	//!< Fat extents in Z
			COMPRESSED	= 0x00010000,	//!< Compressed extents
		};
		enum {
			INVALID		= 0,		//!< Invalid channel format
			R8G8B8		= 1,		//!< 24 bit RGB colour
			R5G6B5		= 2,		//!< 16 bit RGB colour		
			G8			= 3,		//!< 8 bit grayscale
			G16			= 4,		//!< 16 bit grayscale
			L16			= 5,		//!< 16 bit labels
			X8R8G8B8	= 6,		//!< 32 bit XRGB colour
			L16_BIG		= 7,		//!< 16 bit labels (big endian)
			B8G8R8X8	= 8,		//!< 32 bit BGRX colour
		};
		//! Return size of channel elements in bytes
		int channelSize(int s) 
		{ 
			static const int cs[]={0,3,2,1,2,2,4,2,4};
			return cs[s];
		}
		
		//! Name of dataset
		StringA name;
		//! Comments on dataset
		StringA comment;
		//! Origin of dataset
		StringA origin;
		
		//! The stripes composing the extent file
		Stripe stripes[MAXSTRIPES];
		//! Number of stripes in extent file
		Int32 stripeCount;
		//! The source files for this dataset
		FilePattern sources[MAXCHANNELS];
		//! Flags indicating characteristics of extent file
		Int32 flags;
		//! Size of dataset in voxels
		Vector<Int32> datasetSize;
		//! Size of blocks for extent file
		Int32 blockSize;
		//! Number of resolutions in extent file
		Int32 resolutions;
		//! Size of extents
		Vector<Int32> extentSize;
		//! Full size of extents (including fat)
		Vector<Int32> fullExtentSize;
		//! Size of voxels (in arbitrary units, used for aspect ratio)
		Vector<double> voxelSize;
		//! Number of channels
		Int32 channels;
		//! Format of channels
		Int32 channelFormat[MAXCHANNELS];
		//! Name of label source file
		StringA labels;
		
		//! Size of a complete extent in bytes
		Int32 extSize;
		
		//! Offset of a given resolution in file
		int extOff[MAXZOOM];
		//! Number of extents in X direction at a given resolution
		int extX[MAXZOOM];
		//! Number of extents in Y direction at a given resolution
		int extY[MAXZOOM];
		//! Number of extents in Z direction at a given resolution
		int extZ[MAXZOOM];
		//! Size of dataset in voxels in X direction at a given resolution
		int sizeX[MAXZOOM];
		//! Size of dataset in voxels in Y direction at a given resolution
		int sizeY[MAXZOOM];
		//! Size of dataset in voxels in Z direction at a given resolution
		int sizeZ[MAXZOOM];
		
	public:
		//! Constructor
		ExtentFileDescriptor()
		{
			name=""; comment=""; origin="";
			flags=0;
			resolutions=0;
			channels=0;
			stripeCount=0;
			for(int i=0;i<MAXCHANNELS;i++)
				channelFormat[i]=0;
		}
		
		//! Load descriptor from XML
		bool loadFromXML(const char *filename);
		
#ifdef DECLARE_CONVERT
		//! Perform conversion from source file to extent file
		void convert(StripeFunc stripefunc);
#endif
	};
	
}

#endif