/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __FILE_H
#define __FILE_H

#include <windows.h>

namespace extfile
{
	//! Class for encapsulating file access.
	/*! This class is required to allow 64-bit file access. It is OS specific, since
	the standard C library calls do not allow 64-bit offsets within files */
	class FileWrapper
	{
		//! File handle
		HANDLE hFile;
		//! Copy of file name
		char filename[1024];

	public:
		//! Flags for open mode
		enum {
			READ = 1,	//!< Open file for read access
			WRITE = 2	//!< Open file for write access
		};

	public:
		//! Constructor
		FileWrapper() 
		{ 
			hFile=NULL;
		}
		//! Destructor
		~FileWrapper() 
		{ 
			close(); 
		}
		//! Open a file
		/*!
		\param fname Name of file to open
		\param mode Open mode, can be a combination of READ and WRITE
		*/
		bool open(const char *fname, Int32 mode)
		{
			strcpy(filename,fname);
			// If open in write-only, delete file first
			if(mode==WRITE)
				DeleteFile(fname);
			hFile=CreateFile(fname,((mode&READ) ? GENERIC_READ : 0) | ((mode&WRITE) ? GENERIC_WRITE : 0),0,NULL,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
			return hFile!=INVALID_HANDLE_VALUE;
		}
		//! Close the file
		void close(bool destroy=false) 
		{ 
			if(hFile)
			{
				CloseHandle(hFile);
				hFile=NULL;
				if(destroy)
					DeleteFile(filename);
			}
		}
		//! Read from file
		/*!
		\param data Buffer to read into
		\param len Number of bytes to read
		*/
		bool read(void *data, Int32 len)
		{
			DWORD br;
			ReadFile(hFile,data,len,&br,NULL);
			return true;
		}
		//! Write to file
		/*!
		\param data Buffer to write from
		\param len Number of bytes to write
		*/
		bool write(void *data, Int32 len)
		{
			DWORD br;
			WriteFile(hFile,data,len,&br,NULL);
			return true;
		}
		//! Seek within file
		/*!
		\param pos Absolute file position to seek to
		*/
		void seek(Int64 pos)
		{
			union {
				struct { DWORD lp; LONG hp; };
				LONGLONG qp;
			} p;
			p.qp=pos;
			SetFilePointer(hFile,p.lp,&p.hp,FILE_BEGIN);
		}
	};
}


#endif

